\
\ File name: OOP.FS loads on top of gforth
\ 
\ Last_change: KS 02.10.2011 16:53:18
\

Forth definitions  warnings off  hex

\ -----------------------------------------------------------------------------
\ gforth extensions
\ -----------------------------------------------------------------------------

true Constant gforth_062 \ gforth_070 when false

: cell-    1 cells - ;
: ?comp   state @ 0= Abort" compilation only" ;
: ?exec   state @    Abort" execution only" ;

: case?   ( n1 n2 -- n1 ff | tf )   over = dup IF  nip  THEN ;  \ the most primitive case operator

Create restore ] r> r> ! EXIT [
: save  ( var -- )  r> swap dup >r @ >r restore >r >r ;

: becomes  ( xt <name> -- ) \ makes existing <name> behave as xt
   here >r
   >r   ' >body dp !
   postpone ahead   r> >body dp !   postpone THEN
   r> dp !
;
: new.voc ( wid -- )
    dup >r wordlist-struct %size + dup head? -1 = IF ( wid nt )
       dup name>int dup >code-address docon: = swap >body @ r@ = and IF
          id. rdrop EXIT
    THEN THEN
    drop r> body> >head-noprim id. ;
' new.voc becomes .voc

: \\ ( -- )
   SOURCE-ID  1+ 1 U> IF  BEGIN  REFILL 0=  UNTIL THEN
   POSTPONE \
; IMMEDIATE

\ -----------------------------------------------------------------------------
\ general late binding method in gforth, not limited to oop definitions
\ -----------------------------------------------------------------------------
\ >: <name> will compile a branch to the new definition if <name> did exist.
\ Otherwise it will compile a new definition for <name> that starts with
\ a branch to its code.
\ The actual patching of the initial branch at <name> is done by ; later on.
\ Therefore, during the >:-definition, the previous version of <name>
\ is still accessible. This is exploited by EMBED
\ -----------------------------------------------------------------------------


0 Constant #colon    \ magic numbers to
1 Constant #method   \ control ;'s behaviour

: >:  ( <name> -- sys # )
   name   2dup Current @ (search-wordlist)
   ?dup IF  nip nip   name>int >body        \ leave addr of branch to be rewritten
       align docol: cfa,   here             \ leave addr of definitions body
       defstart ] :-hook   #method          \ leave check number
   ELSE                                     \ <name> does not exist yet
      header, (:noname)
      postpone AHEAD  docol: cfa,
      postpone THEN   #colon
   THEN
;
: :       ( -- sys # )   : #colon ;   \ this allows to create an "intelligent" ;

: ;       ( sys # -- )
   #colon  case? IF  postpone ;  EXIT THEN
   #method case? IF  postpone ;
                     here >r
                     >r dp !  postpone AHEAD
                     r> dp !  postpone THEN
                     r> dp !  EXIT
                 THEN
   ABORT" unstructured!"
[ drop ] ; immediate

: Does>   ( # -- # )     >r postpone Does> r> ; immediate

: :noname ( -- # )       :noname #colon ;

: uninitialized ( -- )          true ABORT" un-initialized method" ;

: Method:       ( <name> -- )   >: postpone uninitialized postpone ; ;

\ -----------------------------------------------------------------------------
\ OOP mechanisms on top of gforth
\ -----------------------------------------------------------------------------

Variable LastClass     LastClass off
Variable ClassContext  ClassContext off    \ Current class context

: Forth-context ( -- )  ClassContext off ;

: set-class   ( class -- )  LastClass !  ClassContext on ;

: classorder  ( -- )   \ establish Forth search order for this :-definition
   r> context save >r
   LastClass @ ?dup 0= ABORT" Class context undefined"
   context !
;
Variable (doClass  : doClass ( -- xt )  (doClass @ ;
Variable (doObj    : doObj   ( -- xt )  (doObj @ ;
Variable (doAttr   : doAttr  ( -- xt )  (doAttr @ ;
Variable (doRef    : doRef   ( -- xt )  (doRef @ ;

Variable 'classroot  'classroot off   \ serves as a forward reference

: Class       ( <name> -- )
   Vocabulary  0 , 0 , 'classroot @ ,
\ | search-xt | wordlist | Voclink | ext || attr | size | parent |
DOES>  ( -- )  [ here (doClass ! ] set-class ;

\ -----------------------------------------------------------------------------
\ @Class word set
\ -----------------------------------------------------------------------------
Forth definitions
Vocabulary @Class immediate
@Class definitions

: Context     ( -- cpa )       LastClass ;

: Current     ( -- cpa )
   Current @ cell- @ doClass - ABORT" No class compilation context set"
   Current
;
: '        ( <name> -- )   classorder ' ;

: h'       ( <name> -- )   ' ;                          \ search Forth context. Debugging aid

: c.last   ( cpa -- addr )  @ cell+ ;                   \ addr of vocabulary's wordlist anchor

: c.attr   ( cpa -- addr )  @ [ 4 cells ] Literal + ;
   1 Constant #sealed                                   \ attribute bit
   : ?unsealed   ( cpa -- )
      @Class c.attr @ #sealed and ABORT" Class is sealed"
   ;
: c.size   ( cpa -- addr )  @ [ 5 cells ] Literal + ;

: c.parent ( cpa -- addr )  @ [ 6 cells ] Literal + ;

: Create   ( <name> -- )
   Create @Class LastClass @ , immediate
;
: >object  ( addr -- obj )  dup @ set-class   cell+ ;

: Object   ( <name> -- )
   @Class Create   LastClass c.size @ allot             \ allocate object's data field
DOES> ( -- obj )  [ here (doObj ! ]
   @Class >object
   State @ IF  postpone Literal  THEN
;
: Attribute ( <name> -- )
   @Class Current ?unsealed
   @Class Create      Current c.size @ ,
   LastClass c.size @ Current c.size +!                 \ make room for included object
DOES>  ( obj1 -- obj2 )  [ here (doAttr ! ]
   @Class >object @    ?dup 0= ?EXIT                    \ don't do anything when the offset is 0
   State @ IF  postpone Literal postpone +  EXIT THEN   \ could be optimized by adding to preceeding Literal
   +
;
: dereference  ( obj offset  -- )   + @ ?dup 0= ABORT" un-initialized reference" execute ;

: Reference ( <name> -- )
   @Class Current ?unsealed  [ Forth ]
   @Class Create   Current c.size @ ,                   \ offset into the object. The cell holds an execution token
   @Class  1 cells Current c.size +!                    \ room for execution token of objects or initialization
DOES> ( obj1 -- obj2 )  [ here (doRef ! ]
   @Class >object @
   State @ IF  postpone Literal postpone dereference  EXIT THEN  \ could be optimized by adding to preceeding Literal
   dereference
;
: ==  ( xt u -- xt f )   over cell+ @ = ;

Forth definitions

: BIND     ( xt <proxy name> -- )                       \ bind does not type check, so xt can be anything
   State @ >r   State off                               \ State must be off, because executed objects are state smart
   @Class ' doObj  == 0= ABORT" no object"  execute     \ object containing reference
   BEGIN  dup ' doAttr == WHILE  nip execute  REPEAT    \ skip over potential hierarchy of attributes
   nip   r> State !   Forth-context
   doRef == 0= ABORT" no proxy"   >body cell+ @ +
   State @ IF  postpone Literal postpone !  EXIT THEN
   !
; immediate

\ -----------------------------------------------------------------------------
\ OOP support in the Root wordlist
\ -----------------------------------------------------------------------------
Forth Root definitions Forth

: Self     ( -- )       @Class Current @ set-class ; immediate

: classes  ( -- )  space
   Voclink  BEGIN  @ ?dup
   WHILE  dup cell- cell-  dup body> cell+ @ doClass =
          IF  dup body> >name .name  THEN  drop
   REPEAT
;
: methods  ( -- )   LastClass @ .voc   classorder words ;

Forth definitions
\ -----------------------------------------------------------------------------
\ Methods interpreter/compiler
\ -----------------------------------------------------------------------------

: find-method  ( addr len -- xt | 0 )
   LastClass >r  BEGIN  r@ @
   WHILE  2dup r@ @ (search-wordlist)
          ?dup IF  nip nip rdrop  EXIT THEN
          r> @Class c.parent >r  [ Forth ]
   REPEAT
   rdrop 2drop false
;
\ -----------------------------------------------------------------------------
\ outer interpreter for gforth_0.62
\ -----------------------------------------------------------------------------

: oop-compiler ( c-addr u -- )
   ClassContext @   Forth-context
   IF  2dup find-method ?dup
      IF  nip nip name>comp execute  EXIT
      THEN
      2drop  compiler-notfound EXIT
   THEN
   2dup find-name ?dup
   IF  nip nip name>comp execute  EXIT
   THEN
   2dup snumber? ?dup
   IF  0> IF  swap postpone Literal  THEN
       postpone Literal  2drop  EXIT
   THEN
   compiler-notfound
;
: oop-interpreter ( c-addr u -- )
   ClassContext @
   IF  2dup find-method ?dup
      IF  Forth-context  nip nip name>int execute  EXIT
      THEN
      2drop  interpreter-notfound EXIT
   THEN
   2dup find-name ?dup
   IF  nip nip name>int execute  EXIT
   THEN
   2dup 2>r snumber? IF  2rdrop  EXIT THEN
   2r> interpreter-notfound
;

gforth_062 [IF]

\ these re-definitions cannot be done using >: - I have no idea why

   : new[ ( -- )  ['] oop-interpreter IS parser state off ; immediate
   : new] ( -- )  ['] oop-compiler    IS parser state on  ;
   ' new[ becomes [
   ' new] becomes ]
   [

[ELSE]

cr .( this should hold code for gforth_071 but it doesn't !!! )
\ : host-compiler1  ( addr len -- xt )  host-compiler ['] noop ;
\ ' host-compiler1 IS parser1

[THEN]

: isBranch?  ( x -- f )   \ test x to be a valid branch code
   drop true              \ unfortunately, I don't know how to do this for gforth
;
: embed   ( <name> -- )   \ compile previous version of <name> into <name>'s redefinition
   postpone Self @Class ' >body Forth-context  [ Forth ]
   dup @ isBranch? 0= ABORT" not a method definition"
   cell+ @ body> compile,
; immediate

\ -----------------------------------------------------------------------------
\ ClassRoot and initial methods
\ -----------------------------------------------------------------------------

Class ClassRoot                \ Note: ClassRoot's parent field is set to 0
' ClassRoot >body 'classroot ! \ every class inherits ClassRoot
' ClassRoot >body set-current  ( ClassRoot definitions )

: definitions  ( -- )    LastClass @ Current ! ;

: '      ( <name> -- )   @Class ' ;

: New    ( <name> -- )   @Class Object ;

: ::     ( <name> -- )   @Class Attribute ;

: []     ( <name> -- )   @Class Reference ;

: allot  ( u -- )        @Class Current c.size +! ;

: seal   ( -- )          @Class Current c.attr dup @ #sealed or swap ! ;

: units  ( u1 -- u2 )    @Class Context c.size @ * ;

: inherit ( -- )
   @Class Current c.last @ ABORT" inherit only once"
          Current c.size @ ABORT" use inherit before any attribute"
   LastClass c.size @ Current c.size !
   LastClass        @ Current c.parent !
;
: addr    ( obj -- addr ) Forth-context ; immediate

: ..      ( -- )          Forth-context ; immediate

: size    ( -- bytes )    ?exec LastClass c.size @ ;

: words   ( -- )          methods ;

: methods ( -- )          methods ;

: order   ( -- )  LastClass >r
   BEGIN  r@ @ ?dup WHILE  .voc  r> @Class c.parent >r  REPEAT  rdrop
   ."     "  [ Forth ] Current @ .voc
;
: see     ( <name> -- )   classorder see  ;

include objects.fs

\\

\ -----------------------------------------------------------------------------
\ Late binding by wordlist search, replaced by >: mechanism
\ -----------------------------------------------------------------------------

Variable late-binding  late-binding on

: (lookup  ( obj -- )
   "lit   r> dup cell+ >r   @ set-class
   count find-method Forth-context ?dup 0= ABORT" undefined method"
   name>int execute
;
: lookup  ( <method> -- )
   ?comp   postpone (lookup
   BL parse string, align
   LastClass @ , Forth-context
; immediate

